"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.applyEngineReport = exports.normalizePath = exports.Index = void 0;
const tslib_1 = require("tslib");
const core_1 = require("@yarnpkg/core");
const get_1 = tslib_1.__importDefault(require("lodash/get"));
const set_1 = tslib_1.__importDefault(require("lodash/set"));
const toPath_1 = tslib_1.__importDefault(require("lodash/toPath"));
const unset_1 = tslib_1.__importDefault(require("lodash/unset"));
class Index {
    constructor(indexedFields) {
        this.indexedFields = indexedFields;
        this.items = [];
        this.indexes = {};
        this.clear();
    }
    clear() {
        this.items = [];
        for (const field of this.indexedFields) {
            this.indexes[field] = new Map();
        }
    }
    insert(item) {
        this.items.push(item);
        for (const field of this.indexedFields) {
            const value = Object.prototype.hasOwnProperty.call(item, field)
                ? item[field]
                : undefined;
            if (typeof value === `undefined`)
                continue;
            const list = core_1.miscUtils.getArrayWithDefault(this.indexes[field], value);
            list.push(item);
        }
        return item;
    }
    find(filter) {
        var _a;
        if (typeof filter === `undefined`)
            return this.items;
        const filterEntries = Object.entries(filter);
        if (filterEntries.length === 0)
            return this.items;
        const sequentialFilters = [];
        let matches;
        for (const [field_, value] of filterEntries) {
            const field = field_;
            const index = Object.prototype.hasOwnProperty.call(this.indexes, field)
                ? this.indexes[field]
                : undefined;
            if (typeof index === `undefined`) {
                sequentialFilters.push([field, value]);
                continue;
            }
            const filterMatches = new Set((_a = index.get(value)) !== null && _a !== void 0 ? _a : []);
            if (filterMatches.size === 0)
                return [];
            if (typeof matches === `undefined`) {
                matches = filterMatches;
            }
            else {
                for (const item of matches) {
                    if (!filterMatches.has(item)) {
                        matches.delete(item);
                    }
                }
            }
            if (matches.size === 0) {
                break;
            }
        }
        let result = [...matches !== null && matches !== void 0 ? matches : []];
        if (sequentialFilters.length > 0) {
            result = result.filter(item => {
                for (const [field, value] of sequentialFilters) {
                    const valid = typeof value !== `undefined`
                        ? Object.prototype.hasOwnProperty.call(item, field) && item[field] === value
                        : Object.prototype.hasOwnProperty.call(item, field) === false;
                    if (!valid) {
                        return false;
                    }
                }
                return true;
            });
        }
        return result;
    }
}
exports.Index = Index;
const numberRegExp = /^[0-9]+$/;
const identifierRegExp = /^[a-zA-Z0-9_]+$/;
const knownDictKeys = new Set([`scripts`, ...core_1.Manifest.allDependencies]);
function isKnownDict(parts, index) {
    return index === 1 && knownDictKeys.has(parts[0]);
}
function normalizePath(p) {
    const parts = Array.isArray(p)
        ? p
        : (0, toPath_1.default)(p);
    const normalizedParts = parts.map((part, t) => {
        if (numberRegExp.test(part))
            return `[${part}]`;
        if (identifierRegExp.test(part) && !isKnownDict(parts, t))
            return `.${part}`;
        return `[${JSON.stringify(part)}]`;
    });
    return normalizedParts.join(``).replace(/^\./, ``);
}
exports.normalizePath = normalizePath;
function formatStackLine(configuration, caller) {
    // TODO: Should this be in formatUtils? Might not be super useful as a core feature...
    const parts = [];
    if (caller.methodName !== null)
        parts.push(core_1.formatUtils.pretty(configuration, caller.methodName, core_1.formatUtils.Type.CODE));
    if (caller.file !== null) {
        const fileParts = [];
        fileParts.push(core_1.formatUtils.pretty(configuration, caller.file, core_1.formatUtils.Type.PATH));
        if (caller.line !== null) {
            fileParts.push(core_1.formatUtils.pretty(configuration, caller.line, core_1.formatUtils.Type.NUMBER));
            if (caller.column !== null) {
                fileParts.push(core_1.formatUtils.pretty(configuration, caller.line, core_1.formatUtils.Type.NUMBER));
            }
        }
        parts.push(`(${fileParts.join(core_1.formatUtils.pretty(configuration, `:`, `grey`))})`);
    }
    return parts.join(` `);
}
function applyEngineReport(project, { manifestUpdates, reportedErrors }, { fix } = {}) {
    var _a, _b;
    const changedWorkspaces = new Map();
    const remainingErrors = new Map();
    for (const [workspaceCwd, workspaceUpdates] of manifestUpdates) {
        const workspaceErrors = (_b = (_a = reportedErrors.get(workspaceCwd)) === null || _a === void 0 ? void 0 : _a.map(text => ({ text, fixable: false }))) !== null && _b !== void 0 ? _b : [];
        let changedWorkspace = false;
        const workspace = project.getWorkspaceByCwd(workspaceCwd);
        const manifest = workspace.manifest.exportTo({});
        for (const [fieldPath, newValues] of workspaceUpdates) {
            if (newValues.size > 1) {
                const conflictingValuesMessage = [...newValues].map(([value, sources]) => {
                    const prettyValue = core_1.formatUtils.pretty(project.configuration, value, core_1.formatUtils.Type.INSPECT);
                    const stackLine = sources.size > 0
                        ? formatStackLine(project.configuration, sources.values().next().value)
                        : null;
                    return stackLine !== null
                        ? `\n${prettyValue} at ${stackLine}`
                        : `\n${prettyValue}`;
                }).join(``);
                workspaceErrors.push({ text: `Conflict detected in constraint targeting ${core_1.formatUtils.pretty(project.configuration, fieldPath, core_1.formatUtils.Type.CODE)}; conflicting values are:${conflictingValuesMessage}`, fixable: false });
            }
            else {
                const [[newValue]] = newValues;
                const currentValue = (0, get_1.default)(manifest, fieldPath);
                if (currentValue === newValue)
                    continue;
                if (!fix) {
                    const errorMessage = typeof currentValue === `undefined`
                        ? `Missing field ${core_1.formatUtils.pretty(project.configuration, fieldPath, core_1.formatUtils.Type.CODE)}; expected ${core_1.formatUtils.pretty(project.configuration, newValue, core_1.formatUtils.Type.INSPECT)}`
                        : typeof newValue === `undefined`
                            ? `Extraneous field ${core_1.formatUtils.pretty(project.configuration, fieldPath, core_1.formatUtils.Type.CODE)} currently set to ${core_1.formatUtils.pretty(project.configuration, currentValue, core_1.formatUtils.Type.INSPECT)}`
                            : `Invalid field ${core_1.formatUtils.pretty(project.configuration, fieldPath, core_1.formatUtils.Type.CODE)}; expected ${core_1.formatUtils.pretty(project.configuration, newValue, core_1.formatUtils.Type.INSPECT)}, found ${core_1.formatUtils.pretty(project.configuration, currentValue, core_1.formatUtils.Type.INSPECT)}`;
                    workspaceErrors.push({ text: errorMessage, fixable: true });
                    continue;
                }
                if (typeof newValue === `undefined`)
                    (0, unset_1.default)(manifest, fieldPath);
                else
                    (0, set_1.default)(manifest, fieldPath, newValue);
                changedWorkspace = true;
            }
            if (changedWorkspace) {
                changedWorkspaces.set(workspace, manifest);
            }
        }
        if (workspaceErrors.length > 0) {
            remainingErrors.set(workspace, workspaceErrors);
        }
    }
    return {
        changedWorkspaces,
        remainingErrors,
    };
}
exports.applyEngineReport = applyEngineReport;
